@@ -1,5 +1,8 @@
+import atexit
+import collections
 import datetime
 import base64
+import signal
 from textwrap import dedent
 
 from async_generator import asynccontextmanager
@@ -22,6 +25,7 @@ from .exceptions import (
     CellExecutionError
 )
 from .util import run_sync, ensure_async
+from .output_widget import OutputWidget
 
 
 def timestamp():
@@ -101,21 +105,6 @@ class NotebookClient(LoggingConfigurable):
         ),
     ).tag(config=True)
 
-    nest_asyncio = Bool(
-        False,
-        help=dedent(
-            """
-            If False (default), then blocking functions such as `execute`
-            assume that no event loop is already running. These functions
-            run their async counterparts (e.g. `async_execute`) in an event
-            loop with `asyncio.run_until_complete`, which will fail if an
-            event loop is already running. This can be the case if nbclient
-            is used e.g. in a Jupyter Notebook. In that case, `nest_asyncio`
-            should be set to True.
-            """
-        ),
-    ).tag(config=True)
-
     force_raise_errors = Bool(
         False,
         help=dedent(
@@ -133,7 +122,7 @@ class NotebookClient(LoggingConfigurable):
         ),
     ).tag(config=True)
 
-    extra_arguments = List(Unicode())
+    extra_arguments = List(Unicode()).tag(config=True)
 
     kernel_name = Unicode(
         '',
@@ -291,7 +280,7 @@ class NotebookClient(LoggingConfigurable):
         ----------
         nb : NotebookNode
             Notebook being executed.
-        km : KernerlManager (optional)
+        km : KernelManager (optional)
             Optional kernel manager. If none is provided, a kernel manager will
             be created.
         """
@@ -299,6 +288,15 @@ class NotebookClient(LoggingConfigurable):
         self.nb = nb
         self.km = km
         self.reset_execution_trackers()
+        self.widget_registry = {
+            '@jupyter-widgets/output': {
+                'OutputModel': OutputWidget
+            }
+        }
+        # comm_open_handlers should return an object with a .handle_msg(msg) method or None
+        self.comm_open_handlers = {
+            'jupyter.widget': self.on_comm_open_jupyter_widget
+        }
 
     def reset_execution_trackers(self):
         """Resets any per-execution trackers.
@@ -307,6 +305,11 @@ class NotebookClient(LoggingConfigurable):
         self._display_id_map = {}
         self.widget_state = {}
         self.widget_buffers = {}
+        # maps to list of hooks, where the last is used, this is used
+        # to support nested use of output widgets.
+        self.output_hook_stack = collections.defaultdict(list)
+        # our front-end mimicing Output widgets
+        self.comm_objects = {}
 
     def start_kernel_manager(self):
         """Creates a new kernel manager.
@@ -329,23 +332,21 @@ class NotebookClient(LoggingConfigurable):
         return self.km
 
     async def _async_cleanup_kernel(self):
+        now = self.shutdown_kernel == "immediate"
         try:
-            # Send a polite shutdown request
-            await ensure_async(self.kc.shutdown())
-            try:
-                # Queue the manager to kill the process, sometimes the built-in and above
-                # shutdowns have not been successful or called yet, so give a direct kill
-                # call here and recover gracefully if it's already dead.
-                await ensure_async(self.km.shutdown_kernel(now=True))
-            except RuntimeError as e:
-                # The error isn't specialized, so we have to check the message
-                if 'No kernel is running!' not in str(e):
-                    raise
+            # Queue the manager to kill the process, and recover gracefully if it's already dead.
+            if await ensure_async(self.km.is_alive()):
+                await ensure_async(self.km.shutdown_kernel(now=now))
+        except RuntimeError as e:
+            # The error isn't specialized, so we have to check the message
+            if 'No kernel is running!' not in str(e):
+                raise
         finally:
             # Remove any state left over even if we failed to stop the kernel
             await ensure_async(self.km.cleanup())
-            await ensure_async(self.kc.stop_channels())
-            self.kc = None
+            if getattr(self, "kc"):
+                await ensure_async(self.kc.stop_channels())
+                self.kc = None
 
     _cleanup_kernel = run_sync(_async_cleanup_kernel)
 
@@ -414,6 +415,8 @@ class NotebookClient(LoggingConfigurable):
         When control returns from the yield it stops the client's zmq channels, and shuts
         down the kernel.
         """
+        cleanup_kc = kwargs.pop('cleanup_kc', True)
+
         # Can't use run_until_complete on an asynccontextmanager function :(
         if self.km is None:
             self.start_kernel_manager()
@@ -423,7 +426,8 @@ class NotebookClient(LoggingConfigurable):
         try:
             yield
         finally:
-            self._cleanup_kernel()
+            if cleanup_kc:
+                self._cleanup_kernel()
 
     @asynccontextmanager
     async def async_setup_kernel(self, **kwargs):
@@ -434,20 +438,46 @@ class NotebookClient(LoggingConfigurable):
 
         When control returns from the yield it stops the client's zmq channels, and shuts
         down the kernel.
+
+        Handlers for SIGINT and SIGTERM are also added to cleanup in case of unexpected shutdown.
         """
-        reset_kc = kwargs.pop('reset_kc', False)
+        cleanup_kc = kwargs.pop('cleanup_kc', True)
         if self.km is None:
             self.start_kernel_manager()
 
+        # self._cleanup_kernel uses run_async, which ensures the ioloop is running again.
+        # This is necessary as the ioloop has stopped once atexit fires.
+        atexit.register(self._cleanup_kernel)
+
+        def on_signal():
+            asyncio.ensure_future(self._async_cleanup_kernel())
+            atexit.unregister(self._cleanup_kernel)
+
+        loop = asyncio.get_event_loop()
+        try:
+            loop.add_signal_handler(signal.SIGINT, on_signal)
+            loop.add_signal_handler(signal.SIGTERM, on_signal)
+        except (NotImplementedError, RuntimeError):
+            # NotImplementedError: Windows does not support signals.
+            # RuntimeError: Raised when add_signal_handler is called outside the main thread
+            pass
+
         if not self.km.has_kernel:
             await self.async_start_new_kernel_client(**kwargs)
         try:
             yield
         finally:
-            if reset_kc:
+            if cleanup_kc:
                 await self._async_cleanup_kernel()
 
-    async def async_execute(self, **kwargs):
+            atexit.unregister(self._cleanup_kernel)
+            try:
+                loop.remove_signal_handler(signal.SIGINT)
+                loop.remove_signal_handler(signal.SIGTERM)
+            except (NotImplementedError, RuntimeError):
+                pass
+
+    async def async_execute(self, reset_kc=False, **kwargs):
         """
         Executes each code cell.
 
@@ -457,16 +487,15 @@ class NotebookClient(LoggingConfigurable):
             Any option for `self.kernel_manager_class.start_kernel()`. Because
             that defaults to AsyncKernelManager, this will likely include options
             accepted by `AsyncKernelManager.start_kernel()``, which includes `cwd`.
-            If present, `reset_kc` is passed to `self.async_setup_kernel`:
-            if True, the kernel client will be reset and a new one will be created
-            and cleaned up after execution (default: False).
+
+            `reset_kc` if True, the kernel client will be reset and a new one
+            will be created (default: False).
 
         Returns
         -------
         nb : NotebookNode
             The executed notebook.
         """
-        reset_kc = kwargs.get('reset_kc', False)
         if reset_kc and self.km:
             await self._async_cleanup_kernel()
         self.reset_execution_trackers()
@@ -785,6 +814,14 @@ class NotebookClient(LoggingConfigurable):
     def output(self, outs, msg, display_id, cell_index):
         msg_type = msg['msg_type']
 
+        parent_msg_id = msg['parent_header'].get('msg_id')
+        if self.output_hook_stack[parent_msg_id]:
+            # if we have a hook registered, it will overrride our
+            # default output behaviour (e.g. OutputWidget)
+            hook = self.output_hook_stack[parent_msg_id][-1]
+            hook.output(outs, msg, display_id, cell_index)
+            return
+
         try:
             out = output_from_msg(msg)
         except ValueError:
@@ -810,6 +847,15 @@ class NotebookClient(LoggingConfigurable):
 
     def clear_output(self, outs, msg, cell_index):
         content = msg['content']
+
+        parent_msg_id = msg['parent_header'].get('msg_id')
+        if self.output_hook_stack[parent_msg_id]:
+            # if we have a hook registered, it will overrride our
+            # default clear_output behaviour (e.g. OutputWidget)
+            hook = self.output_hook_stack[parent_msg_id][-1]
+            hook.clear_output(outs, msg, cell_index)
+            return
+
         if content.get('wait'):
             self.log.debug('Wait to clear output')
             self.clear_before_next_output = True
@@ -830,6 +876,23 @@ class NotebookClient(LoggingConfigurable):
             self.widget_state.setdefault(content['comm_id'], {}).update(data['state'])
             if 'buffer_paths' in data and data['buffer_paths']:
                 self.widget_buffers[content['comm_id']] = self._get_buffer_data(msg)
+        # There are cases where we need to mimic a frontend, to get similar behaviour as
+        # when using the Output widget from Jupyter lab/notebook
+        if msg['msg_type'] == 'comm_open':
+            target = msg['content'].get('target_name')
+            handler = self.comm_open_handlers.get(target)
+            if handler:
+                comm_id = msg['content']['comm_id']
+                comm_object = handler(msg)
+                if comm_object:
+                    self.comm_objects[comm_id] = comm_object
+            else:
+                self.log.warning(f'No handler found for comm target {target!r}')
+        elif msg['msg_type'] == 'comm_msg':
+            content = msg['content']
+            comm_id = msg['content']['comm_id']
+            if comm_id in self.comm_objects:
+                self.comm_objects[comm_id].handle_msg(msg)
 
     def _serialize_widget_state(self, state):
         """Serialize a widget state, following format in @jupyter-widgets/schema."""
@@ -854,6 +917,33 @@ class NotebookClient(LoggingConfigurable):
             )
         return encoded_buffers
 
+    def register_output_hook(self, msg_id, hook):
+        """Registers an override object that handles output/clear_output instead.
+
+        Multiple hooks can be registered, where the last one will be used (stack based)
+        """
+        # mimics
+        # https://jupyterlab.github.io/jupyterlab/services/interfaces/kernel.ikernelconnection.html#registermessagehook
+        self.output_hook_stack[msg_id].append(hook)
+
+    def remove_output_hook(self, msg_id, hook):
+        """Unregisters an override object that handles output/clear_output instead"""
+        # mimics
+        # https://jupyterlab.github.io/jupyterlab/services/interfaces/kernel.ikernelconnection.html#removemessagehook
+        removed_hook = self.output_hook_stack[msg_id].pop()
+        assert removed_hook == hook
+
+    def on_comm_open_jupyter_widget(self, msg):
+        content = msg['content']
+        data = content['data']
+        state = data['state']
+        comm_id = msg['content']['comm_id']
+        module = self.widget_registry.get(state['_model_module'])
+        if module:
+            widget_class = module.get(state['_model_name'])
+            if widget_class:
+                return widget_class(comm_id, state, self.kc, self)
+
 
 def execute(nb, cwd=None, km=None, **kwargs):
     """Execute a notebook's code, updating outputs within the notebook object.
